from distutils.core import setup
setup(
  name = 'py2gmsh',
  packages = ['py2gmsh'],
  version = '0.1.1',
  description = 'Python wrappers to create a gmsh geo file',
  author = 'Tristan de Lataillade',
  author_email = 'delataillade.tristan@gmail.com',
  url = 'https://github.com/tridelat/py2gmsh',
  download_url = 'https://github.com/tridelat/py2gmsh/tarball/0.1',
  keywords = ['gmsh', 'wrapper', 'mesh'],
  classifiers = [],
)
